/* finddomain.c -- handle list of needed message catalogs
   Copyright (C) 1995 Software Foundation, Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <errno.h>
#include <stdio.h>
#include <sys/types.h>

#ifdef __GNUC__
# define alloca __builtin_alloca
#else
# if HAVE_ALLOCA_H 
#  include <alloca.h>
# else
#  ifdef _AIX
 #pragma alloca
#  else
#   ifndef alloca
char *alloca ();
#   endif
#  endif
# endif
#endif

#ifdef STDC_HEADERS
# include <stdlib.h>
#else
char *getenv ();
# ifdef HAVE_MALLOC_H
#  include <malloc.h>
# else
void free ();
# endif
#endif

#if HAVE_STRING_H
# include <string.h>
#else
# include <strings.h>
#endif
#if !HAVE_STRCHR
# ifndef strchr
#  define strchr index
# endif
#endif

#include "gettext.h"
#include "gettextP.h"
#include "libgettext.h"

#ifndef errno
extern int errno;
#endif

/* @@ end of prolog @@ */


#ifndef HAVE_GETCWD
char *getwd ();
# define getcwd(buf, max) getwd (buf)
#else
char *getcwd ();
#endif

/* Amount to increase buffer size by in each try.  */
#define PATH_INCR 32

/* The following is from pathmax.h.  */
/* Non-POSIX BSD systems might have gcc's limits.h, which doesn't define
   PATH_MAX but might cause redefinition warnings when sys/param.h is
   later included (as on MORE/BSD 4.3).  */
#if defined(_POSIX_VERSION) || (defined(HAVE_LIMITS_H) && !defined(__GNUC__))
# include <limits.h>
#endif

#ifndef _POSIX_PATH_MAX
# define _POSIX_PATH_MAX 255
#endif

#if !defined(PATH_MAX) && defined(_PC_PATH_MAX)
# define PATH_MAX (pathconf ("/", _PC_PATH_MAX) < 1 ? 1024 : pathconf ("/", _PC_PATH_MAX))
#endif

/* Don't include sys/param.h if it already has been.  */
#if defined(HAVE_SYS_PARAM_H) && !defined(PATH_MAX) && !defined(MAXPATHLEN)
# include <sys/param.h>
#endif

#if !defined(PATH_MAX) && defined(MAXPATHLEN)
# define PATH_MAX MAXPATHLEN
#endif

#ifndef PATH_MAX
# define PATH_MAX _POSIX_PATH_MAX
#endif

/* Contains the default location of the message catalogs.  */
const char _nl_default_dirname[] = GNULOCALEDIR;

/* List with bindings of specific domains created by bindtextdomain()
   calls.  */
struct binding *_nl_domain_bindings;

/* List of already loaded domains.  */
extern struct loaded_domain *_nl_loaded_domains;

/* Prototypes for library functions.  */
void error __P ((int, int, const char *, ...));
void *xmalloc __P ((size_t __n));

/* Prototypes for local functions.  */
static const char *category_to_name __P((int category));
#if !defined HAVE_SETLOCALE || !defined HAVE_LC_MESSAGES || !defined HAVE_LOCALE_NULL
static const char *guess_category_value __P((int category,
					     const char *categoryname));
#endif


/* Return a data structure describing the message catalog described by
   the DOMAINNAME and CATEGORY parameters with respect to the currently
   established bindings.  */
struct loaded_domain *
_nl_find_domain (domainname, category)
     const char *domainname;
     int category;
{
  const char *categoryname;
  const char *categoryvalue;
  struct binding *binding;
  char *dirname;
  char *filename;
  size_t filename_len;
  struct loaded_domain *retval;

  /* Now determine the symbolic name of CATEGORY and its value.  */
  categoryname = category_to_name (category);
#if defined HAVE_SETLOCALE && defined HAVE_LOCALE_NULL
# ifndef HAVE_LC_MESSAGES
  if (category == LC_MESSAGES)
    categoryvalue = guess_category_value (category, categoryname);
  else
# endif
    categoryvalue = setlocale (category, NULL);
#else
  categoryvalue = guess_category_value (category, categoryname);
#endif

  /* If the current locale value is C (or POSIX) we don't load a domain.
     The causes the calling function to return the MSGID.  */ 
  if (strcmp (categoryvalue, "C") == 0 || strcmp (categoryvalue, "POSIX") == 0)
    return NULL;

  /* First find matching binding.  */
  for (binding = _nl_domain_bindings; binding != NULL; binding = binding->next)
    {
      int compare = strcmp (domainname, binding->domainname);
      if (compare == 0)
	/* We found it!  */
	break;
      if (compare < 0)
	{
	  /* It is not in the list.  */
	  binding = NULL;
	  break;
	}
    }

  if (binding == NULL)
    dirname = (char *) _nl_default_dirname;
  else if (binding->dirname[0] == '/')
    dirname = binding->dirname;
  else
    {
      /* We have a relative path.  Make it absolute now.  */
      size_t dirname_len = strlen (binding->dirname) + 1;
      size_t path_max;
      char *ret;

      path_max = (unsigned) PATH_MAX;
      path_max += 2;		/* The getcwd docs say to do this.  */

      dirname = (char *) alloca (path_max + dirname_len);

      errno = 0;
      while ((ret = getcwd (dirname, path_max)) == NULL && errno == ERANGE)
	{
	  path_max += PATH_INCR;
	  dirname = (char *) alloca (path_max + dirname_len);
	  errno = 0;
	}

      if (ret == NULL)
	error (1, errno, "while getting current directory");

      stpcpy (stpcpy (strchr (dirname, '\0'), "/"), binding->dirname);
    }

  /* Now we have all components which describe the message catalog to be
     used.  Create an absolute file name.  */
  filename_len = strlen (dirname) + 1 + strlen (categoryvalue) + 1
		 + strlen (categoryname) + 1 + strlen (domainname) + 4;
  filename = (char *) xmalloc (filename_len);
  sprintf (filename, "%s/%s/%s/%s.mo", dirname, categoryvalue,
	   categoryname, domainname);

  /* Look in list of already loaded file whether it is already available.  */
  for (retval = _nl_loaded_domains; retval != NULL; retval = retval->next)
    {
      int compare = strcmp (retval->filename, filename);
      if (compare == 0)
	/* We found it!  */
	break;
      if (compare < 0)
	{
	  /* It's not in the list.  */
	  retval = NULL;
	  break;
	}
    }

  /* If it is not in list try to load it.  Free the allocated FILENAME
     if it is not used.  */
  if (retval != NULL || (retval = _nl_load_msg_cat (filename)) == NULL)
    free (filename);

  return retval;
}

/* Return string representation of locale CATEGORY.  */
static const char *category_to_name (category)
     int category;
{
  const char *retval;

  switch (category)
  {
#ifdef LC_COLLATE
  case LC_COLLATE:
    retval = "LC_COLLATE";
    break;
#endif
#ifdef LC_CTYPE
  case LC_CTYPE:
    retval = "LC_CTYPE";
    break;
#endif
#ifdef LC_MONETARY
  case LC_MONETARY:
    retval = "LC_MONETARY";
    break;
#endif
#ifdef LC_NUMERIC
  case LC_NUMERIC:
    retval = "LC_NUMERIC";
    break;
#endif
#ifdef LC_TIME
  case LC_TIME:
    retval = "LC_TIME";
    break;
#endif
#ifdef LC_MESSAGES
  case LC_MESSAGES:
    retval = "LC_MESSAGES";
    break;
#endif
#ifdef LC_RESPONSE
  case LC_RESPONSE:
    retval = "LC_RESPONSE";
    break;
#endif
#ifdef LC_ALL
  case LC_ALL:
    /* This might not make sense but is perhaps better than any other
       value.  */
    retval = "LC_ALL";
    break;
#endif
  default:
    /* If you have a better idea for a default value let me know.  */
    retval = "LC_XXX";
  }

  return retval;
}

#if !defined HAVE_SETLOCALE || !defined HAVE_LC_MESSAGES || !defined HAVE_LOCALE_NULL
/* Guess value of current locale from value of the environment variables.  */
static const char *guess_category_value (category, categoryname)
     int category;
     const char *categoryname;
{
  const char *retval;

  /* Setting of LC_ALL overwrites all other.  */
  retval = getenv ("LC_ALL");
  if (retval != NULL && retval[0] != '\0')
    return retval;

  /* Next comes the name of the desired category.  */
  retval = getenv (categoryname);
  if (retval != NULL && retval[0] != '\0')
    return retval;

  /* Last possibility is the LANG environment variable.  */
  retval = getenv ("LANG");
  if (retval != NULL && retval[0] != '\0')
    return retval;

  /* We use C as the default domain.  POSIX says this is implementation
     defined.  */
  return "C";
}
#endif
