/* gettext - retrieve text string from message catalog and print it.
   Copyright (C) 1995 Free Software Foundation, Inc.
   Written by Ulrich Drepper <drepper@gnu.ai.mit.edu>, May 1995.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <getopt.h>
#include <stdio.h>

#ifdef STDC_HEADERS
# include <stdlib.h>
#else
char *getenv ();
#endif

#ifdef HAVE_LOCALE_H
# include <locale.h>
#endif

#include "error.h"
#include "system.h"

#include <libintl.h>
#include "libgettext.h"

#define _(str) gettext (str)

/* Name the program is called with.  */
char *program_name;

/* Long options.  */
static const struct option long_options[] =
{
  { "help", no_argument, NULL, 'h' },
  { "version", no_argument, NULL, 'V' },
  { NULL, 0, NULL, 0 }
};

/* Prototypes for local functions.  */
static void usage __P ((int status))
#if defined __GNUC__ && ((__GNUC__ == 2 && __GNUC_MINOR__ >= 5) || __GNUC__ > 2)
     __attribute__ ((noreturn))
#endif
;

int
main (argc, argv)
     int argc;
     char *argv[];
{
  int optchar;
  int do_help = 0;
  int do_version = 0;
  char *msgid;
  char *domain = getenv ("TEXTDOMAIN");
  char *domaindir = getenv ("TEXTDOMAINDIR");

  /* Set program name for message texts.  */
  program_name = argv[0];

#ifdef HAVE_SETLOCALE
  /* Set locale via LC_ALL.  */
  setlocale (LC_ALL, "");
#endif

  /* Set the text message domain.  */
  bindtextdomain (PACKAGE, LOCALEDIR);
  textdomain (PACKAGE);

  while ((optchar = getopt_long (argc, argv, "hV", long_options, NULL)) != EOF)
    switch (optchar)
    {
    case '\0':
      break;
    case 'h':
      do_help = 1;
      break;
    case 'V':
      do_version = 1;
      break;
    default:
      usage (EXIT_FAILURE);
    }

  /* Version information is requested.  */
  if (do_version)
    {
      fprintf (stderr, "%s - GNU %s %s\n", program_name, PACKAGE, VERSION);
      exit (EXIT_SUCCESS);
    }

  /* Help is requested.  */
  if (do_help)
    usage (EXIT_SUCCESS);
      
  if (optind >= argc)
    error (EXIT_FAILURE, 0, _("missing arguments"));

  /* Get arguments.  */
  msgid = argv[optind++];
  if (optind < argc)
    {
      domain = msgid;
      msgid = argv[optind++];

      if (optind < argc)
	error (EXIT_FAILURE, 0, _("too many arguments"));
    }

  /* If no domain name is given we print the original string.  */
  if (domain == NULL || domain[0] == '\0')
    {
      fputs (msgid, stdout);
      exit (EXIT_SUCCESS);
    }

  /* Bind domain to appropriate directory.  */
  if (domaindir != NULL && domaindir[0] != '\0')
    __bindtextdomain (domain, domaindir);

  /* Write out the result.  */
  fputs (__dgettext (domain, msgid), stdout);

  exit (EXIT_SUCCESS);
}


/* Display usage information and exit.  */
static void
usage (status)
     int status;
{
  if (status != EXIT_SUCCESS)
    fprintf (stderr, _("Try `%s --help' for more information\n"),
	     program_name);
  else
    printf (_("\
Usage: %s [-h | -v | [TEXTDOMAIN] MSGID]\n\
  -h, --help           display this help and exit\n\
  -v, --version        display version information and exit\n\
  [TEXTDOMAIN] MSGID   retrieve translated message corresponding\n\
                       to MSGID from TEXTDOMAIN\n\
\n\
If the TEXTDOMAIN parameter is not given, the domain is determined from the\n\
environment variable TEXTDOMAIN.  If the message catalog is not found in the\n\
regular directory, another location can be specified with the environment\n\
variable TEXTDOMAINDIR.\n\
Standard search directory: %s\n"), program_name, LOCALEDIR);

  exit (status);
}
