/* hash - implement simple hashing table with string based keys.
   Copyright (C) 1994, 1995 Software Foundation, Inc.
   Written by Ulrich Drepper <drepper@gnu.ai.mit.edu>, October 1994.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>
#include <sys/types.h>

#if HAVE_OBSTACK
# include <obstack.h>
#else
# include "obstack.h"
#endif

#ifdef HAVE_VALUES_H
# include <values.h>
#endif

#include "hash.h"
#include "system.h"

#define obstack_chunk_alloc xmalloc
#define obstack_chunk_free free

#ifndef BITSPERBYTE
# define BITSPERBYTE 8
#endif

#ifndef	LONGBITS
# define LONGBITS (sizeof (long) * BITSPERBYTE)
#endif

#ifndef bcopy
# define bcopy(s, d, n)	memcpy ((d), (s), (n))
#endif	

void *xmalloc __P ((size_t __n));

typedef struct hash_entry
  {
    unsigned int used;
    char *key;
    void *data;
    struct hash_entry *next;
  }
hash_entry;

/* Prototypes for local functions.  */
static size_t lookup __P ((hash_table *htab, const char *key, size_t keylen,
			   unsigned long hval));
static unsigned long compute_hashval __P ((const char *key, size_t keylen));
static int is_prime __P ((unsigned long candidate));


int
init_hash(htab, init_size)
     hash_table *htab;
     unsigned long init_size;
{
  /* We need the size to be a prime.  */
  init_size = next_prime (init_size);

  /* Initialize the data structure.  */
  htab->size = init_size;
  htab->filled = 0;
  htab->first = NULL;
  htab->table = (void *) calloc (init_size + 1, sizeof (hash_entry));
  obstack_init (&htab->mem_pool);

  return htab->table == NULL;
}


int
delete_hash(htab)
     hash_table *htab;
{
  free (htab->table);
  obstack_free (&htab->mem_pool, NULL);
  return 0;
}


int
insert_entry (htab, key, keylen, data)
     hash_table *htab;
     const char *key;
     size_t keylen;
     void *data;
{
  unsigned long hval = compute_hashval (key, keylen);
  hash_entry *table = (hash_entry *) htab->table;
  size_t idx = lookup (htab, key, keylen, hval);

  if (table[idx].used)
    /* We don't want to overwrite the old value.  */
    return 1;
  else
    {
      /* An empty bucket has been found.  */
      table[idx].used = hval;
      table[idx].key = obstack_copy0 (&htab->mem_pool, key, keylen);
      table[idx].data = data;

      /* List the new value in the list.  */
      if ((hash_entry *) htab->first == NULL)
	{
	  table[idx].next = &table[idx];
	  *(hash_entry **) &htab->first = &table[idx];
	}
      else
	{
	  table[idx].next = ((hash_entry *) htab->first)->next;
	  ((hash_entry *) htab->first)->next = &table[idx];
	  *(hash_entry **) &htab->first = &table[idx];
	}

      ++htab->filled;
      if (100 * htab->filled > 90 * htab->size)
	{
	  /* Resize the table.  */
	  unsigned long old_size = htab->size;

	  htab->size = next_prime (htab->size * 2);
	  htab->filled = 0;
	  htab->first = NULL;
	  htab->table = (void *) calloc (htab->size, sizeof (hash_entry));

	  for (idx = 1; idx <= old_size; ++idx)
	    if (table[idx].used)
	      insert_entry (htab, table[idx].key, strlen(table[idx].key),
			    table[idx].data);

	  free (table);
	}
      return 0;
    }
  /* NOTREACHED */
}


int
find_entry (htab, key, keylen, result)
     hash_table *htab;
     const char *key;
     size_t keylen;
     void **result;
{
  hash_entry *table = (hash_entry *) htab->table;
  size_t idx = lookup (htab, key, keylen, compute_hashval (key, keylen));
  int retval;

  retval = table[idx].used;
  *result = retval ? table[idx].data : NULL;

  return retval;
}


int
iterate_table (htab, ptr, key, data)
     hash_table *htab;
     void **ptr;
     void **key;
     void **data;
{
  if (*ptr == NULL)
    {
      if (htab->first == NULL)
	return 0;
      *ptr = (void *) ((hash_entry *) htab->first)->next;
    }
  else
    {
      if (*ptr == htab->first)
	return 0;
      *ptr = (void *) (((hash_entry *) *ptr)->next);
    }

  *key = ((hash_entry *) *ptr)->key;
  *data = ((hash_entry *) *ptr)->data;
  return 1;
}


static size_t
lookup (htab, key, keylen, hval)
     hash_table *htab;
     const char *key;
     size_t keylen;
     unsigned long hval;
{
  unsigned long hash;
  size_t idx;
  hash_entry *table = (hash_entry *) htab->table;

  /* First hash function: simply take the modul but prevent zero.  */
  hash = 1 + hval % htab->size;

  idx = hash;
  
  if (table[idx].used)
    {
      if (table[idx].used == hval && table[idx].key[keylen] == '\0'
	  && strncmp (key, table[idx].key, keylen) == 0)
	return idx;

      /* Second hash function as suggested in [Knuth].  */
      hash = 1 + hval % (htab->size - 2);

      do
	{
	  if (idx <= hash)
	    idx = htab->size + idx - hash;
	  else
	    idx -= hash;

	  /* If entry is found use it.  */
	  if (table[idx].used == hval && table[idx].key[keylen] == '\0'
	      && strncmp (key, table[idx].key, keylen) == 0)
	    return idx;
	}
      while (table[idx].used);
    }
  return idx;
}


static unsigned long
compute_hashval (key, keylen)
     const char *key;
     size_t keylen;
{
  size_t cnt;
  unsigned long hval, g;
  /* Compute the hash value for the given string.  */
  cnt = 0;
  hval = keylen;
  while (cnt < keylen)
    {
      hval <<= 4;
      hval += key[cnt++];
      g = hval & ((unsigned long) 0xf << (LONGBITS - 4));
      if (g != 0)
	{
	  hval ^= g >> (LONGBITS - 8);
	  hval ^= g;
	}
    }
  return hval != 0 ? hval : ~((unsigned long) 0);
}


unsigned long
next_prime (seed)
     unsigned long seed;
{
  /* Make it definitely odd.  */
  seed |= 1;

  while (!is_prime (seed))
    seed += 2;

  return seed;
}


static int
is_prime (candidate)
     unsigned long candidate;
{
  /* No even number and none less than 10 will be passed here.  */
  unsigned long div = 3;
  unsigned long sq = div * div;

  while (sq < candidate && candidate % div != 0)
    {
      ++div;
      sq += 4 * div;
      ++div;
    }

  return candidate % div != 0;
}
