/* msgfmt - converts Uniforum style .mo file to binary .po file
   Copyright (C) 1995 Free Software Foundation, Inc.
   Written by Ulrich Drepper <drepper@gnu.ai.mit.edu>, April 1995.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <ctype.h>
#include <errno.h>
#include <getopt.h>
#include <stdio.h>
#include <sys/param.h>
#include <sys/types.h>

#ifdef STDC_HEADERS
# include <stdlib.h>
#endif

#ifdef HAVE_LOCALE_H
# include <locale.h>
#endif

#include "hash.h"

#include "error.h"
#include "getline.h"
#include "printf.h"
#include <system.h>

#include "gettext.h"
#include "domain.h"
#include "hash-string.h"
#include <libintl.h>

#define _(str) gettext (str)

#ifndef errno
extern int errno;
#endif

/* Define the data structure which we need to represent the data to
   be written out.  */
struct id_str_pair
{
  char *id;
  char *str;
};

/* Contains information about the definition of one translation.  */
struct msgstr_def
{
  char *msgstr;
  const char *file_name;
  size_t line_no;
};

/* Alignment of strings in resulting .mo file.  */
static size_t alignment;

/* Search path list for input files.  */
char *input_path_list;

/* Nonzero if no hash table in .mo is wanted.  */
static int no_hash_table;

/* Nonzero if an output file is opened.  */
static int output_file_open;

/* Specifies name of the output file.  */
static const char *output_file_name;

/* String containing name the program is called with.  */
const char *program_name;

/* We may have more than one input file.  Domains with same names in
   different files have to merged.  So we need a list of tables for
   each output file.  */
static struct msg_domain *domain;
static struct msg_domain *current_domain;

/* If not zero list duplicate message identifiers.  */
static int verbose;

/* Long options.  */
static const struct option long_options[] =
{
  { "alignment", required_argument, NULL, 'a' },
  { "help", no_argument, NULL, 'h' },
  { "input-path", required_argument, NULL, 'I' },
  { "no-hash", no_argument, &no_hash_table, 1 },
  { "output-file", required_argument, NULL, 'o' },
  { "verbose", no_argument, &verbose, 1 },
  { "version", no_argument, NULL, 'V' },
  { NULL, 0, NULL, 0 }
};


#ifndef roundup
# if defined __GNUC__ && __GNUC__ >= 2
#  define roundup(x, y) ({typeof(x) _x = (x); typeof(y) _y = (y); \
			  ((_x + _y - 1) / _y) * _y; })
# else
#  define roundup(x, y) ((((x)+((y)-1))/(y))*(y))
# endif	/* GNU CC2  */
#endif /* roundup  */

/* Prototypes for local functions.  */
static void usage __P ((int status))
#if defined __GNUC__ && ((__GNUC__ == 2 && __GNUC_MINOR__ >= 5) || __GNUC__ > 2)
	__attribute__ ((noreturn))
#endif
;
static void error_print ();
static struct msg_domain *new_domain __P ((const char *name));
static void process_po_file __P ((FILE *input_file,
				  const char *input_file_name));
static int compare_id __P ((const void *pval1, const void *pval2));
static void write_table __P ((FILE *output_file, hash_table *tab));
static void check_pair __P ((const char *file_name,
			     const char *msgid, size_t msgid_line,
			     const char *msgstr, size_t msgstr_line));


int
main(argc, argv)
     int argc;
     char *argv[];
{
  int optchar;
  int do_help = 0;
  int do_version = 0;

  /* Set default value for global variables.  */
  alignment = DEFAULT_ALIGNMENT;

  /* Set program name for messages.  */
  program_name = argv[0];
  error_print_progname = error_print;

#ifdef HAVE_SETLOCALE
  /* Set locale via LC_ALL.  */
  setlocale (LC_ALL, "");
#endif

  /* Set the text message domain.  */
  bindtextdomain (PACKAGE, LOCALEDIR);
  textdomain (PACKAGE);

  while ((optchar = getopt_long (argc, argv, "a:hI:o:vV", long_options, NULL))
	 != EOF)
    switch (optchar)
      {
      case '\0':
	break;
      case 'a':
	{
	  char *endp;
	  size_t new_align = strtoul (optarg, &endp, 0);

	  if (endp != optarg)
	    alignment = new_align;
	}
	break;
      case 'h':
	do_help = 1;
	break;
      case 'I':
	if (input_path_list == NULL)
	  input_path_list = xstrdup (optarg);
	else
	  {
	    size_t len = strlen (input_path_list);

	    input_path_list = (char *) xrealloc (input_path_list,
						 len + strlen (optarg) + 2);
	    stpcpy (stpcpy (&input_path_list[len], ":"), optarg);
	  }
	break;
      case 'o':
	output_file_name = optarg;
	break;
      case 'v':
	verbose = 1;
	break;
      case 'V':
	do_version = 1;
	break;
      default:
	usage (EXIT_FAILURE);
	break;
      }

  /* Complete the input file path list with the system default list.  */
  if (input_path_list == NULL)
    input_path_list = STD_INC_PATH;
  else
    {
      size_t len = strlen (input_path_list);
      input_path_list = (char *) xrealloc (input_path_list,
					   len + sizeof (STD_INC_PATH) + 2);
#if __STDC__
      strcpy (&input_path_list[len], ":" STD_INC_PATH);
#else
      stpcpy (stpcpy (&input_path_list[len], ":"), STD_INC_PATH);
#endif	/* not standard C compiler.  */
    }

  /* Version information is requested.  */
  if (do_version)
    {
      fprintf (stderr, "%s - GNU %s %s\n", program_name, PACKAGE, VERSION);
      exit (EXIT_SUCCESS);
    }

  /* Help is requested.  */
  if (do_help)
    usage (EXIT_SUCCESS);

  /* Test whether we have an .po file name as argument.  */
  if (optind >= argc)
    {
      error (EXIT_SUCCESS, 0, _("no input file given"));
      usage (EXIT_FAILURE);
    }

  /* The -o option determines the name of the domain and therefor
     the output file.  */
  if (output_file_name != NULL)
    current_domain = new_domain (output_file_name);

  /* Process all given .po files.  */
  while (argc > optind)
    {
      char *input_file_name;
      FILE *input_file;

      /* If no output file is given the output of the messages upto the first
	 `domain' keyword is placed for each .po file in `messages.mo'.  */
      if (output_file_name == NULL)
	output_file_open = 0;

      /* Open the input file.  Handle `-' as name for stdin.  Please note
	 that me don't free the input_file_name because it is referenced
	 in the constructed tables.  */
      input_file = open_po_file (argv[optind], &input_file_name, 1);
      if (input_file == NULL)
	error (EXIT_FAILURE, errno,
	       _("error while opening \"%s\" for reading"), input_file_name);

      /* Remember that we currently have not specified any domain.  This
	 is of course not true when we saw the -o option.  */
      if (output_file_name == NULL)
	current_domain = NULL;
      /* And process the input file.  */
      process_po_file (input_file, input_file_name);

      /* Prepare for next round.  */
      if (input_file != stdin)
	fclose (input_file);

      ++optind;
    }

  /* Now write out all domains.  */
  while (domain != NULL)
    {
      FILE *output_file;

      /* If no entry for this domain don't even create the file.  */
      if (domain->symbol_tab.filled != 0)
	{
	  if (strcmp (domain->domain_name, "-") == 0)
	    output_file = stdout;
	  else
	    {
	      output_file = fopen (domain->domain_name, "w");
	      if (output_file == NULL)
		error (EXIT_FAILURE, errno,
		       _("error while opening \"%s\" for writing"),
		       domain->domain_name);
	    }

	  write_table (output_file, &domain->symbol_tab);
	  if (output_file != stdout)
	    fclose (output_file);
	}

      domain = domain->next;
    }

  exit (EXIT_SUCCESS);
}


/* Display usage information and exit.  */
static void
usage (status)
     int status;
{
  if (status != EXIT_SUCCESS)
    fprintf (stderr, _("Try `%s --help' for more information\n"),
	     program_name);
  else
    printf (_("\
Usage: %s [OPTION] filename.po ...\n\
Mandatory arguments to long options are mandatory for short options too.\n\
  -a, --alignment=NUMBER   align strings to NUMBER bytes (default: %d)\n\
  -h, --help               display this help and exit\n\
  -I, --input-path=LIST    list of directories searched for input files\n\
      --no-hash            binary file will not include the hash table\n\
  -o, --output-file=FILE   specify output file name as FILE\n\
  -v, --verbose            list input file anomalies\n\
  -V, --version            output version information and exit\n\
\n\
Search path is: %s\n\
\n\
If input file is -, standard input is read.  If output file is -,\n\
output is written to standard output.\n"),
	    program_name, DEFAULT_ALIGNMENT, input_path_list);

  exit (status);
}


static struct msg_domain *
new_domain (name)
     const char *name;
{
  struct msg_domain **p_dom = &domain;

  while (*p_dom != NULL && strcmp (name, (*p_dom)->domain_name) != 0)
    p_dom = &(*p_dom)->next;

  if (*p_dom == NULL)
    {
      *p_dom = (struct msg_domain *) xmalloc (sizeof (**p_dom));

      if (init_hash (&(*p_dom)->symbol_tab, 100) != 0)
	error (EXIT_FAILURE, errno, _("while creating hash table"));
      (*p_dom)->domain_name = xstrdup (name);
      (*p_dom)->next = NULL;
    }

  return *p_dom;
}


/* The address of this function will be assigned to the hook in the error
   functions.  */
static void
error_print ()
{
  /* We don't want the program name to be printed in messages.  */
}


/* Read and process the content of a single .po file.  */
static void
process_po_file (input_file, input_file_name)
     FILE *input_file;
     const char *input_file_name;
{
  static char *buffer = NULL;
  static size_t buffer_size = 0;
  static char *strbuf = NULL;
  static size_t strbuf_size = 0;
  char *msgid_string = NULL;
  size_t msgid_line = 0;
  size_t strbuf_ptr;
  size_t line_no = 1;
  int line_len = getline (&buffer, &buffer_size, input_file);
  char *cp = buffer;

  if (line_len < 0)
    /* End of file is now reached.  */
    return;

  if (buffer == NULL)
    error (EXIT_FAILURE, errno, _("error while reading \"%s\""),
	   input_file_name);

  /* And mark end of line.  */
  buffer[line_len] = '\0';

  while (line_len > 0)
    {
      char *startp, *directive;

      /* Skip leading white spaces.  */
      while (isspace (*cp))
	++cp;

      /* Check for comment line.  */
      if (*cp == '#')
	{
	  /* A comment line might span across many lines when terminated
	     by "\\\n".  */
	  do
	    {
	      line_len = getline (&buffer, &buffer_size, input_file);
	      if (line_len < 0)
		break;
	      ++line_no;
	    }
	  while (buffer[line_len - 2] == '\\' && buffer[line_len - 1] == '\n');
	  continue;
	}

      if (*cp == '\0')
	{
	  /* The line was empty.  */
	  line_len = getline (&buffer, &buffer_size, input_file);
	  if (buffer == NULL)
	    error (EXIT_FAILURE, errno, _("error while reading \"%s\""),
		   input_file_name);

	  cp = buffer;
	  continue;
	}

      /* Now it's clear: the line has to be a directive.  These are
	 seperated by a white space from the rest of the line.  */
      startp = cp;
      while (*cp != '\0' && !isspace (*cp))
	++cp;
      directive = (char *) xmalloc (cp - startp + 1);
      memcpy (directive, startp, cp - startp);
      directive[cp - startp] = '\0';

      while (1)
	{
	  /* Skip separating white spaces.  */
	  while (isspace (*cp))
	    ++cp;

	  if ((*cp != '\\' || *(cp + 1) != '\n') && *cp != '\0')
	    break;

	  line_len = getline (&buffer, &buffer_size, input_file);
	  if (line_len < 0)
	    {
	      error (0, 0, _("incomplete `%s' directive at end of file"),
		     directive);
	      return;
	    }
	  ++line_no;
	  buffer[line_len] = '\0';
	  cp = buffer;
	}

      /* Get the argument string.  The argument may extend on more than
	 one line, following the same syntax as strings in C.  */
      if (*cp != '"')
	{
	  error_with_loc (0, 0, input_file_name, line_no,
			  _("syntax error: value is no quoted string"));
	  continue;
	}
      strbuf_ptr = 0;
      do
	{
	  /* Skip " character.  */
	  ++cp;

	  while (*cp != '\0' && *cp != '"')
	    {
	      if (strbuf_ptr >= strbuf_size)
		{
		  strbuf_size = 2 * (strbuf_size + 100);
		  strbuf = xrealloc (strbuf, strbuf_size);
		}

	      if (*cp == '\\')
		switch (*++cp)
		  {
		  case '\n':
		    /* The value continues on the next line.  */
		    line_len = getline (&buffer, &buffer_size, input_file);
		    if (line_len < 0)
		      {
			error (0, 0,
			       _("end of file \"%s\" while reading string"),
			       input_file_name);
			line_len = 0;
		      }
		    else
		      ++line_no;
		    buffer[line_len] = '\0';
		    cp = buffer;
		    break;
		  case 'n':
		    strbuf[strbuf_ptr++] = '\n';
		    ++cp;
		    break;
		  case 't':
		    strbuf[strbuf_ptr++] = '\t';
		    ++cp;
		    break;
		  case 'v':
		    strbuf[strbuf_ptr++] = '\v';
		    ++cp;
		    break;
		  case 'b':
		    strbuf[strbuf_ptr++] = '\b';
		    ++cp;
		    break;
		  case 'r':
		    strbuf[strbuf_ptr++] = '\r';
		    ++cp;
		    break;
		  case 'f':
		    strbuf[strbuf_ptr++] = '\f';
		    ++cp;
		    break;
		  case '\\':
		    strbuf[strbuf_ptr++] = '\\';
		    ++cp;
		    break;
		  case '"':
		    strbuf[strbuf_ptr++] = '"';
		    ++cp;
		    break;
		  case '0': case '1': case '2': case '3':
		  case '4': case '5': case '6': case '7':
		    {
		      int val = *cp++ - '0';
		      if ('0' <= *cp && *cp <= '7')
			{
			  val *= 8;
			  val += *cp++ - '0';
			  if ('0' <= *cp && *cp <= '7')
			    {
			      val *= 8;
			      val += *cp++ - '0';
			    }
			}
		      strbuf[strbuf_ptr++] = val;
		    }
		    break;
		  case 'x': case 'X':
		    if (isxdigit (*++cp))
		      {
			int val = isdigit (*cp) ? *cp - '0'
						: tolower (*cp) - 'a' + 10;
			if (isxdigit (*++cp))
			  {
			    val *= 16;
			    val += isdigit (*cp) ? *cp - '0'
						 : tolower (*cp) - 'a' + 10;
			    ++cp;
			  }
			strbuf[strbuf_ptr++] = val;
		      }
		    else
		      error_with_loc (0, 0, input_file_name, line_no,
				      _("illegal control sequence"));
		    break;
		  case '\0':
		    error (EXIT_FAILURE, 0, _("\
end of file occured inside directive value"));
		    /* NOTREACHED */
		  default:
		    error_with_loc (0, 0, input_file_name, line_no,
				    _("illegal control sequence"));
		  }
	      else
		strbuf[strbuf_ptr++] = *cp++;
	    }

	  if (*cp == '\0')
	    error_with_loc (0, 0, input_file_name, line_no,
			    _("unterminated string constant"));
	  else
	    {
	      ++cp;
	      while (isspace (*cp))
		++cp;

	      if (*cp != '\0' && *cp != '"')
		{
		  error_with_loc (0, 0, input_file_name, line_no,
				  _("trailing garbage in line"));
		  *cp = '\0';
		}

	      while (line_len > 0 && *cp == '\0')
		{
		  line_len = getline (&buffer, &buffer_size, input_file);
		  if (line_len < 0)
		    line_len = 0;
		  else
		    ++line_no;
		  buffer[line_len] = '\0';
		  cp = buffer;

		  while (isspace (*cp))
		    ++cp;
		}
	    }
	}
      while (*cp == '"');
      if (strbuf_ptr >= strbuf_size)
	{
	  strbuf_size = 2 * (strbuf_size + 100);
	  strbuf = xrealloc (strbuf, strbuf_size);
	}
      strbuf[strbuf_ptr] = '\0';

      /* Test for the known directives and process them or give an error.  */
      if (strcmp (directive, "domain") == 0)
	{
	  /* If no output file was given we change it with each domain
	     directive.  */
	  if (msgid_string != NULL)
	    {
	      error_with_loc (0, 0, input_file_name, line_no, _("\
previous `msgid' directive not matched by `msgstr'"));
	      msgid_string = NULL;
	    }

	  if (output_file_name == NULL)
	    {
	      /* Do some tests on the file name (white space etc).  */
	      size_t correct = strcspn (strbuf, INVALID_PATH_CHAR);

	      if (strbuf[correct] != '\0')
		{
		  error (0, 0, _("\
domain name \"%s\" not suitable as file name: will use prefix"),
			 strbuf);
		  strbuf[correct] = '\0';
		}

	      /* Set new domain.  */
	      current_domain = new_domain (strbuf);
	    }
	}
      else if (strcmp (directive, "msgstr") == 0)
	{
	  if (msgid_string == NULL)
	    error_with_loc (0, 0, input_file_name, line_no,
			    _("`msgid' directive expected"));
	  else if (strbuf[0] == '\0')
	    {
	      if (verbose)
		error_with_loc (0, 0, input_file_name, line_no,
				_("empty `msgstr' entry ignored"));

	      free (msgid_string);
	      msgid_string = NULL;
	    }
	  else
	    {
	      /* We found a valid pair of msgid/msgstr.  Construct struct
		 to decribe msgstr definition.  */
	      struct msgstr_def *entry =
		  (struct msgstr_def *) xmalloc (sizeof (*entry));

	      entry->msgstr = xstrdup (strbuf);
	      entry->file_name = input_file_name;
	      entry->line_no = line_no;

	      /* If --verbose is selected we do some more checks on the
		 both strings.  */
	      if (verbose)
		check_pair (input_file_name, msgid_string, msgid_line,
			    strbuf, line_no);

	      /* Check whether already a domain is specified.  If not use
		 default domain.  */
	      if (current_domain == NULL)
		current_domain = new_domain ("messages");

	      /* We can insert the ID/string pair into the hashing table.  But
		 we have to take care for duplicates.  */
	      if (insert_entry (&current_domain->symbol_tab, msgid_string,
				strlen (msgid_string), entry) != 0
		  && verbose)
		{
		  /* We don't need the just constructed entry.  */
		  free (entry);

		  /* Get the old definition.  */
		  find_entry (&current_domain->symbol_tab, msgid_string,
			      strlen (msgid_string), (void **) &entry);

		  error_with_loc (0, 0, input_file_name, line_no,
				  _("duplicate message ID"));
		  error_with_loc (0, 0, entry->file_name, entry->line_no, _("\
this is the location of the first definition"));
		}

	      free (msgid_string);
	      msgid_string = NULL;
	    }
	}
      else if (strcmp (directive, "msgid") == 0)
	{
	  if (msgid_string != NULL)
	    error_with_loc (0, 0, input_file_name, line_no,
			    _("`msgstr' directive expected"));

	  msgid_string = xstrdup (strbuf);
	  msgid_line = line_no;
	}
      else
	error_with_loc (0, 0, input_file_name, line_no,
			_("syntax error: unknown directive"));

      /* Now we don't need the directive name anymore.  */
      free (directive);
    }

  /* Test for an incompleted msgid/msgstr pair at the end of file.  */
  if (msgid_string != NULL)
    error_with_loc (0, 0, input_file_name, line_no,
		    _("incomplete msgid/msgstr pair at end of file"));
}


static int
compare_id (pval1, pval2)
     const void *pval1;
     const void *pval2;
{
  return strcmp (((struct id_str_pair *)pval1)->id,
		 ((struct id_str_pair *)pval2)->id);
}


static void
write_table (output_file, tab)
     FILE *output_file;
     hash_table *tab;
{
  static char null = '\0';
  /* This should be explained:
     Each string has an associate hashing value V, computed by a fixed
     function.  To locate the string we use open addressing with double
     hashing.  The first index will be V % M, where M is the size of the
     hashing table.  If no entry is found, iterating with a second,
     independent hashing function takes place.  This second value will
     be 1 + V % (M - 2).
     The approximate number of probes will be

       for unsuccessful search:  (1 - N / M) ^ -1
       for successful search:    - (N / M) ^ -1 * ln (1 - N / M)

     where N is the number of keys.

     If we now choose M to be the next prime bigger than 4 / 3 * N,
     we get the values
			 4   and   1.85  resp.
     Because unsuccesful searches are unlikely this is a good value.
     Formulas: [Knuth, The Art of Computer Programming, Volume 3,
		Sorting and Searching, 1973, Addison Wesley]  */
  nls_uint32 hash_tab_size = no_hash_table ? 0
					   : next_prime ((tab->filled * 4)
							 / 3);
  nls_uint32 *hash_tab;

  /* Header of the .mo file to be written.  */
  struct mo_file_header header;
  struct id_str_pair *msg_arr;
  void *ptr;
  size_t cnt;
  char *id;
  struct msgstr_def *entry;
  struct string_desc sd;

  /* Fill the structure describing the header.  */
  header.magic = _MAGIC;		/* Magic number.  */
  header.revision = MO_REVISION_NUMBER;	/* Revision number of file format.  */
  header.nstrings = tab->filled;	/* Number of strings.  */
  header.orig_tab_offset = sizeof (header);
			/* Offset of table for original string offsets.  */
  header.trans_tab_offset = sizeof (header)
			    + tab->filled * sizeof (struct string_desc);
			/* Offset of table for translation string offsets.  */
  header.hash_tab_size = hash_tab_size;	/* Size of used hashing table.  */
  header.hash_tab_offset =
	no_hash_table ? 0 : sizeof (header)
			    + 2 * (tab->filled * sizeof (struct string_desc));
			/* Offset of hashing table.  */

  /* Write the header out.  */
  fwrite (&header, sizeof (header), 1, output_file);

  /* Allocate table for the all elements of the hashing table.  */
  msg_arr = (struct id_str_pair *) alloca (tab->filled * sizeof (msg_arr[0]));

  /* Read values from hashing table into array.  */
  for (cnt = 0, ptr = NULL;
       iterate_table (tab, &ptr, (void **) &id, (void **) &entry) != 0; ++cnt)
    {
      msg_arr[cnt].id = id;
      msg_arr[cnt].str = entry->msgstr;
    }

  /* Sort the table according to original string.  */
  qsort (msg_arr, tab->filled, sizeof (msg_arr[0]), compare_id);

  /* Set offset to first byte after all the tables.  */
  sd.offset = roundup (sizeof (header)
		       + tab->filled * sizeof (sd)
		       + tab->filled * sizeof (sd)
		       + hash_tab_size * sizeof (nls_uint32),
		       alignment);

  /* Write out length and starting offset for all original strings.  */
  for (cnt = 0; cnt < tab->filled; ++cnt)
    {
      sd.length = strlen (msg_arr[cnt].id);
      fwrite (&sd, sizeof (sd), 1, output_file);
      sd.offset += roundup (sd.length + 1, alignment);
    }

  /* Write out length and starting offset for all translation strings.  */
  for (cnt = 0; cnt < tab->filled; ++cnt)
    {
      sd.length = strlen (msg_arr[cnt].str);
      fwrite (&sd, sizeof (sd), 1, output_file);
      sd.offset += roundup (sd.length + 1, alignment);
    }

  /* Skip this part when no hash table is needed.  */
  if (!no_hash_table)
    {
      /* Allocate room for the hashing table to be written out.  */
      hash_tab = (nls_uint32 *) alloca (hash_tab_size * sizeof (nls_uint32));
      memset (hash_tab, '\0', hash_tab_size * sizeof (nls_uint32));

      /* Insert all value in the hash table, following the algorithm described
	 above.  */
      for (cnt = 0; cnt < tab->filled; ++cnt)
	{
	  nls_uint32 hash_val = hash_string (msg_arr[cnt].id);
	  nls_uint32 idx = hash_val % hash_tab_size;

	  if (hash_tab[idx] != 0)
	    {
	      /* We need the second hashing function.  */
	      nls_uint32 c = 1 + (hash_val % (hash_tab_size - 2));

	      do
		if (idx >= hash_tab_size - c)
		  idx -= hash_tab_size - c;
		else
		  idx += c;
	      while (hash_tab[idx] != 0);
	    }

	  hash_tab[idx] = cnt + 1;
	}

      /* Write the hash table out.  */
      fwrite (hash_tab, sizeof (nls_uint32), hash_tab_size, output_file);
    }

  /* Write bytes to make first string to be aligned.  */
  cnt = sizeof (header) + 2 * tab->filled * sizeof (sd)
	+ hash_tab_size * sizeof (nls_uint32);
  fwrite (&null, 1, roundup (cnt, alignment) - cnt, output_file);

  /* Now write the original strings.  */
  for (cnt = 0; cnt < tab->filled; ++cnt)
    {
      size_t len = strlen (msg_arr[cnt].id);

      fwrite (msg_arr[cnt].id, len + 1, 1, output_file);
      fwrite (&null, 1, roundup (len + 1, alignment) - (len + 1), output_file);
    }

  /* Now write the translation strings.  */
  for (cnt = 0; cnt < tab->filled; ++cnt)
    {
      size_t len = strlen (msg_arr[cnt].str);

      fwrite (msg_arr[cnt].str, len + 1, 1, output_file);
      fwrite (&null, 1, roundup (len + 1, alignment) - (len + 1), output_file);

      free (msg_arr[cnt].str);
    }

  /* Hashing table is not used anmore.  */
  delete_hash (tab);
}

static void
check_pair (file_name, msgid, msgid_line, msgstr, msgstr_line)
     const char *file_name;
     const char *msgid;
     size_t msgid_line;
     const char *msgstr;
     size_t msgstr_line;
{
  size_t msgid_len = strlen (msgid);
  size_t msgstr_len = strlen (msgstr);
  size_t nidfmts, nstrfmts;

  /* If the msgid string is empty we have the special entry reserved for
     information about the translation.  */
  if (msgid_len == 0)
    return;

  /* Test 1: check whether both or none of the strings begin with a '\n'.  */
  if ((msgid[0] == '\n') ^ (msgstr[0] == '\n') != 0)
    error_with_loc (0, 0, file_name, msgid_line, _("\
`msgid' and `msgstr' entry does not both begin with '\\n'"));

  /* Test 2: check whether both or none of the strings end with a '\n'.  */
  if ((msgid[msgid_len - 1] == '\n') ^ (msgstr[msgstr_len - 1] == '\n') != 0)
    error_with_loc (0, 0, file_name, msgid_line, _("\
`msgid' and `msgstr' entry does not both end with '\\n'"));

  /* Test 3: check whether both formats strings contain the same number
     of format specifications.  */
  nidfmts = parse_printf_format (msgid, 0, NULL);
  nstrfmts = parse_printf_format (msgstr, 0, NULL);
  if (nidfmts != nstrfmts)
    error_with_loc (0, 0, file_name, msgid_line, _("\
number of format specifications in `msgid' and `msgstr' does not match"));
  else
    {
      int *id_args = (int *) alloca (nidfmts * sizeof (int));
      int *str_args = (int *) alloca (nstrfmts * sizeof (int));
      size_t cnt;

      (void) parse_printf_format (msgid, nidfmts, id_args);
      (void) parse_printf_format (msgstr, nstrfmts, str_args);

      for (cnt = 0; cnt < nidfmts; ++cnt)
	if (id_args[cnt] != str_args[cnt])
	  error_with_loc (0, 0, file_name, msgid_line, _("\
format specifications for argument %u are not the same"), cnt);
    }
}
