/* xgettext - extracts strings from C source file to Uniforum style .po file
   Copyright (C) 1995 Free Software Foundation, Inc.
   Written by Ulrich Drepper <drepper@gnu.ai.mit.edu>, April 1995.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <ctype.h>
#include <errno.h>
#include <getopt.h>
#include <pwd.h>
#include <stdio.h>
#include <time.h>
#include <sys/types.h>

#ifdef STDC_HEADERS
# include <stdlib.h>
#endif

#ifdef HAVE_LOCALE_H
# include <locale.h>
#endif

#ifdef HAVE_UNISTD_H
# include <unistd.h>
#endif

#ifndef errno
extern int errno;
#endif

#include "error.h"
#include "hash.h"
#include "getline.h"
#include "system.h"

#include "gettext.h"
#include "domain.h"
#include <libintl.h>

#ifndef _POSIX_VERSION
struct passwd *getpwuid ();
#endif

/* A convenience macro.  I don't like writing gettext() every time.  */
#define _(str) gettext (str)


/* If nonzero add all comments immediately preceding one of the keywords. */
static int add_all_comments;

/* If nonzero add comments for file name and line number for each msgid.  */
static int line_comment;

/* Tag used in comment of prevailing domain.  */
static char *comment_tag;

/* If nonzero also parse C++ comments.  */
static int cplusplus_comments;

/* Name of default domain file.  If not set defaults to messages.po.  */
static char *default_domain;

/* Content of .po files with symbols to be excluded.  */
static struct msg_domain *exclude_symbols;

/* If nonzero extract all strings.  */
static int extract_all;

/* String containing names of all input files.  */
static char *input_file_names;

/* Search path list for input files.  */
char *input_path_list;

/* If nonzero join existing .po file.  */
static int join_existing;

/* List of keywords to be looked for.  If none given the 3 default keywords
   gettext, dgettext, and dcgettext are used.  */
static size_t nkeywords;
static size_t max_keywords;
struct keyword
{
  size_t len;
  const char *str;
};
static struct keyword *keywords;

/* If nonzero don't use default keywords.  */
static int no_default_keywords;

/* String used as prefix for msgstr.  */
static char *msgstr_prefix;

/* String used as suffix for msgstr.  */
static char *msgstr_suffix;

/* Directory in which output files are created.  */
static char *output_dir;

/* The symbol tables which contain the produced msgid/msgstr pairs.  */
static struct msg_domain *new_symbols;
static struct msg_domain *current;

/* If nonzero omit header with information about this run.  */
static int omit_header;

/* String containing name the program is called with.  */
const char *program_name;

/* If nonzero sort output by msgids.  Duplicates are removed.  */
static int sort_output;

/* If nonzero generate strict Uniforum style .po file.  */
static strict_uniforum;

/* If nonzero be verbose in giving warnings.  */
static int verbose;

/* String length from with on warning are given for possible problem
   while exceeding tools limits.  */
static size_t warn_id_len;

/* String with option xgettext is invoked with.  */
static char *xgettext_options;

/* Long options.  */
static const struct option long_options[] =
{
  { "add-location", no_argument, &line_comment, 1 },
  { "add-comments", optional_argument, NULL, 'c' },
  { "c++", no_argument, &cplusplus_comments, 1 },
  { "default-domain", required_argument, NULL, 'd' },
  { "exclude-file", required_argument, NULL, 'x' },
  { "extract-all", no_argument, &extract_all, 1 },
  { "help", no_argument, NULL, 'h' },
  { "input-path", required_argument, NULL, 'I' },
  { "no-location", no_argument, &line_comment, 0 },
  { "join-existing", no_argument, &join_existing, 1 },
  { "keyword", optional_argument, NULL, 'k' },
  { "msgstr-prefix", optional_argument, NULL, 'm' },
  { "msgstr-suffix", optional_argument, NULL, 'M' },
  { "omit-header", no_argument, &omit_header, 1 },
  { "output-dir", required_argument, NULL, 'p' },
  { "sort-output", no_argument, &sort_output, 1 },
  { "strict", no_argument, &strict_uniforum, 1 },
  { "string-limit", required_argument, NULL, 'l' },
  { "verbose", no_argument, NULL, 'v' },
  { "version", no_argument, NULL, 'V' },
  { NULL, 0, NULL, 0 }
};

/* Hold file name and number of the line where the string occurs.  */
struct file_name_line
{
  const char *file_name;
  size_t line_no;
  char *comment;
  struct file_name_line *next;
};

/* Structure used while sorting messages.  */
struct id_str
{
  char *msgid;
  char *msgstr;
  struct file_name_line occurence;
};

/* Description of a dynamically extended buffer.  */
struct line_buffer
{  /* Address of buffer.  */
   char *buf;
   /* Current size of buffer.  */
   size_t max_size;
   /* Currently used size.  */
   size_t act_size;
   /* Line number.  */
   size_t number;
   /* Name of the file.  */
   const char *file_name;
};
/* These variables provide two buffers which are allocated and extended by
   getline().  */
static struct line_buffer line = { NULL, 0 };
static struct line_buffer string = { NULL, 0 };
static struct line_buffer comment = { NULL, 0 };


/* Prototypes for local functions.  */
static void usage __P ((int status))
#if defined __GNUC__ && ((__GNUC__ == 2 && __GNUC_MINOR__ > 4) || __GNUC__ > 2)
	__attribute__ ((noreturn))
#endif
;
static void error_print ();
static void process_c_source __P ((FILE *c_file, const char *file_name));
static void skip_c_char __P ((char **cp));
static void process_c_string __P ((FILE *c_file, char **cp));
static void process_comment __P ((FILE *c_file, char **cp));
static void process_gettext __P ((FILE *c_file, char **cp, int arg_no));

static struct msg_domain *read_po_file __P ((const char *file_name,
					     struct msg_domain *msgs,
					     const char *force_domain,
					     int use_path));
static struct msg_domain * new_domain __P ((struct msg_domain **po_content,
					    const char *domain_name));

static void add_id_str __P ((struct msg_domain *dom, const char *msgid,
			     const char *c_file_name, size_t line_no));
static void write_out_domain __P ((struct msg_domain *dom, int do_sort));
static int compare_msgid __P ((const void *arg1, const void *arg2));
static long difftm __P ((struct tm *__a, struct tm *__b));
static void write_header __P ((FILE *output_file));

int
main (argc, argv)
     int argc;
     char *argv[];
{
  size_t act_len, max_len, line_len;
  int cnt;
  int optchar;
  int do_help = 0;
  int do_version = 0;

  /* Set program name for messages.  */
  program_name = argv[0];
  error_print_progname = error_print;
  warn_id_len = WARN_ID_LEN;

#ifdef HAVE_SETLOCALE
  /* Set locale via LC_ALL.  */
  setlocale (LC_ALL, "");
#endif

  /* Set the text message domain.  */
  bindtextdomain (PACKAGE, LOCALEDIR);
  textdomain (PACKAGE);

  /* Set initial value of variables.  */
  line_comment = 1;
  input_path_list = xstrdup (STD_INC_PATH);
  default_domain = "messages";
  exclude_symbols = new_domain (&exclude_symbols, "exclude");

  while ((optchar = getopt_long (argc, argv, "ac::Cd:hI:jk::l:m::M::np:svVx:",
				 long_options, NULL)) != EOF)
    switch (optchar)
      {
      case '\0':
	break;
      case 'a':
	extract_all = 1;
	break;
      case 'c':
	if (optarg == NULL)
	  {
	    add_all_comments = 1;
	    comment_tag = NULL;
	  }
	else
	  {
	    add_all_comments = 0;
	    comment_tag = optarg;
	    /* We ignore leading white space.  */
	    while (isspace (*comment_tag))
	      ++comment_tag;
	  }
	break;
      case 'C':
	cplusplus_comments = 1;
	break;
      case 'd':
	default_domain = optarg;
	break;
      case 'h':
	do_help = 1;
	break;
      case 'I':
	/* You may wonder why I did it this complicated way.  The reason is
	   that we may read .po files during processing the command line
	   options and so this path list must already be intact.  Especially
	   the default system path should be available.  */
	{
	  size_t len = strlen (input_path_list);
	  size_t alen = strlen (optarg);

	  input_path_list = xrealloc (input_path_list, len + alen + 2);

	  memmove (&input_path_list[len - sizeof (STD_INC_PATH) + alen + 2],
		   &input_path_list[len - sizeof (STD_INC_PATH) + 1],
		   len + 1);
	  strcpy (&input_path_list[len - sizeof (STD_INC_PATH) + 1], optarg);
	  input_path_list[len - sizeof (STD_INC_PATH) + 1 + alen] = ':';
	}
	break;
      case 'j':
	join_existing = 1;
	break;
      case 'k':
	if (optarg == NULL)
	  no_default_keywords = 1;
	else
	  if (optarg[0] != '\0')
	    {
	      if (nkeywords >= max_keywords)
		{
		  max_keywords += 10;
		  keywords = (struct keyword *)
		    xrealloc (keywords, max_keywords * sizeof (*keywords));
		}
	      keywords[nkeywords].len = strlen (optarg);
	      keywords[nkeywords].str = optarg;
	      ++nkeywords;
	    }
	break;
      case 'l':
	{
	  char *endp;
	  size_t tmp_val = strtoul (optarg, &endp, 0);
	  if (endp[0] == '\0')
	    warn_id_len = tmp_val;
	}
	break;
      case 'm':
	/* -m takes an optional argument.  If none is given "" is assumed. */
	msgstr_prefix = optarg == NULL ? "" : optarg;
	break;
      case 'M':
	/* -M takes an optional argument.  If none is given "" is assumed. */
	msgstr_suffix = optarg == NULL ? "" : optarg;
	break;
      case 'n':
	line_comment = 1;
	break;
      case 'p':
	{
	  size_t len = strlen (optarg);

	  if (output_dir != NULL)
	    free (output_dir);

	  if (optarg[len - 1] == '/')
	    output_dir = xstrdup (optarg);
	  else
	    {
	      asprintf (&output_dir, "%s/", optarg);
	      if (output_dir == NULL)
		/* We are about to construct the absolute path to the
		   directory for the output files but asprintf failed.  */
		error (EXIT_FAILURE, errno, _("while preparing output"));
	    }
	}
	break;
      case 's':
	sort_output = 1;
	break;
      case 'v':
	verbose = 1;
	break;
      case 'V':
	do_version = 1;
	break;
      case 'x':
	exclude_symbols = read_po_file (optarg, exclude_symbols,
					"exclude", 1);
	break;
      default:
	usage (EXIT_FAILURE);
	/* NOTREACHED */
      }

  /* Normalize selected options.  */
  if (omit_header != 0)
    line_comment = 0;

  /* Version information requested.  */
  if (do_version)
    {
      fprintf (stderr, "%s - GNU %s %s\n", program_name, PACKAGE, VERSION);
      exit (EXIT_SUCCESS);
    }

  /* Help is requested.  */
  if (do_help)
    usage (EXIT_SUCCESS);

  /* Test whether we have some input files given.  */
  if (optind >= argc)
    {
      error (EXIT_SUCCESS, 0, _("no input file given"));
      usage (EXIT_FAILURE);
    }

  /* Canonize msgstr prefix/suffix.  */
  if (msgstr_prefix != NULL && msgstr_suffix == NULL)
    msgstr_suffix = "";
  else if (msgstr_prefix == NULL && msgstr_suffix != NULL)
    msgstr_prefix = NULL;

  /* Default output directory is the current directory.  */
  if (output_dir == NULL)
    output_dir = "./";

  /* We now create the default domain.  This will always be the first in
     the linked list.  */
  current = new_domain (&new_symbols, default_domain);

  /* Compose a string with all command line options.  */
  max_len = act_len = 0;
  for (cnt = 1; cnt < optind; ++cnt)
    {
      size_t len = strlen (argv[cnt]);
      if (act_len + len + 1 >= max_len)
	{
	  max_len += MAX ((unsigned) 100, 2 * len);
	  xgettext_options = xrealloc (xgettext_options, max_len);
	}
      xgettext_options[act_len++] = ' ';
      strcpy (&xgettext_options[act_len], argv[cnt]);
      act_len += len;
    }

  /* We need a string with the names of all .c files.  */
  max_len = act_len = line_len = 0;
  for (cnt = optind; cnt < argc; ++cnt)
    {
      size_t len;
      char *cp = argv[cnt];

      /* Remove leading "./".  */
      while (cp[0] == '.' && cp[1] == '/')
	cp += 2;

      /* Now we can get the string length.  */
      len = strlen (cp);
      if (act_len + len + 5 >= max_len)
	{
	  max_len += MAX ((unsigned) 100, 2 * len);
	  input_file_names = xrealloc (input_file_names, max_len);
	}
      if (line_len > 0 && line_len + len > 65)
	{
	  input_file_names[act_len++] = '\n';
	  line_len = 0;
	}
      input_file_names[act_len++] = ' ';
      strcpy (&input_file_names[act_len], cp);
      act_len += len;
      line_len += len;
    }

  /* Process all input files.  */
  do
    {
      /* Locate input file and open it.  */
      if (strcmp (argv[optind], "-") == 0)
	process_c_source (stdin, "<stdin>");
      else
	{
	  FILE *input_file = fopen (argv[optind], "r");
	  if (input_file == NULL)
	    error (EXIT_FAILURE, errno,
		   _("error while opening \"%s\" for reading"), argv[optind]);

	  process_c_source (input_file, argv[optind]);

	  fclose (input_file);
	}
      ++optind;
    }
  while (optind < argc);

  /* If sorting was requested sort and write out.  In any case close
     the file afterwards.  */
  for (current = new_symbols; current != NULL; current = current->next)
    /* Write out the table after possibly sorting it.  */
    write_out_domain (current, sort_output);

  exit (EXIT_SUCCESS);
}


/* Display usage information and exit.  */
static void
usage (status)
     int status;
{
  if (status != EXIT_SUCCESS)
    fprintf (stderr, _("Try `%s --help' for more information\n"),
	     program_name);
  else
    {
      printf (_("\
Usage: %s [OPTION] INPUTFILE ...\n\
Mandatory arguments to long options are mandatory for short options too.\n\
  -a, --extract-all              extract all strings\n\
  -c, --add-comments[=TAG]       place comment block with TAG (or those\n\
                                 preceding keyword lines) in output file\n\
  -C, --c++                      recognize C++ style comments\n\
  -d, --default-domain=NAME      use NAME.po for output (instead of messages.po)\n\
  -h, --help                     display this help and exit\n\
  -I, --input-path=LIST          list of directories searched for input files\n\
  -j, --join-existing            join messages with existing file\n\
  -k, --keyword[=WORD]           additonal keyword to be looked for (without\n\
                                 WORD means not to use default keywords)\n\
  -l, --string-limit=NUMBER      set string length limit to NUMBER instead %u\n\
  -m, --msgstr-prefix[=STRING]   use STRING or \"\" as prefix for msgstr entries\n\
  -M, --msgstr-suffix[=STRING]   use STRING or \"\" as suffix for msgstr entries\n\
      --no-location              do not write '#: filename:line' lines\n"),
	      program_name, WARN_ID_LEN);
      printf (_("\
  -n, --add-location             generate '#: filename:line' lines (default)\n\
      --omit-header              don't write header with `msgid \"\"' entry\n\
  -p, --output-dir=DIR           output files will be placed in directory DIR\n\
  -s, --sort-output              generate sorted output and remove duplicates\n\
      --strict                   write out strict Uniforum conforming .po file\n\
  -v, --verbose                  be verbose when giving warnings\n\
  -V, --version                  output version information and exit\n\
  -x, --exclude-file=FILE        entries from FILE are not extracted\n\
\n\
Search path for supplementary .po files is:\n\
%s\n\
\n\
If INPUTFILE is -, standard input is read.\n"), input_path_list);
    }

  exit (status);
}


/* The address of this function will be assigned to the hook in the error
   functions.  */
static void
error_print ()
{
  /* We don't want the program name to be printed in messages.  */
}


/* Read the INPUT_FILE as a C source and extract strings.  */
static void
process_c_source (c_file, c_file_name)
     FILE *c_file;
     const char *c_file_name;
{
  comment.act_size = 0;
  line.number = 0;
  /* Set current file name.  Remove leading "./".  */
  line.file_name = c_file_name;
  while (line.file_name[0] == '.' && line.file_name[1] == '/')
    line.file_name += 2;

  while (!feof (c_file))
    {
      /* We are only interested in
	 a) comments
	 b) strings and character constants
	 c) {,d,dc}gettext function calls  */
      char *cp;
      int len;

      /* If we are expected to copy all comments in the same line or the
	 one before we have to check if the last gathered one may be
	 expired.  */
      if (add_all_comments
	  && comment.act_size != 0 && comment.number < line.number)
	comment.act_size = 0;

      len = getline (&line.buf, &line.max_size, c_file);
      if (len < 0)
	return;

      ++line.number;
      line.buf[len] = '\0';
      line.act_size = (size_t) len;
      cp = line.buf;

      while (isspace (*cp))
	++cp;

      if (*cp == '#')
	{
	  /* The line contains a preprocessor directive.  */
	  while (isspace (*++cp));

	  /* As a special case we have to deal with the #line directive.  */
	  if (!isalnum (cp[4]) && cp[4] != '_' && strncmp (cp, "line", 4) == 0)
	    {
	      /* We have a problem here:  because we are not running
		 a preprocessor we cannot expand the form
		     #line pp-tokens new-line
		 We simply ignore it here as we do for all other illegal
		 forms.  */
	      size_t new_line_num;
	      char *endp;

	      cp += 4;
	      while (isspace (cp[0]))
		++cp;

	      /* Here should come a digit-sequence.  */
	      new_line_num = strtoul (cp, &endp, 0);
	      if (cp == endp)
		/* No number found: ignore the whole line.  */
		continue;

	      line.number = new_line_num - 1;

	      /* Read upto next argument or end of line.  */
	      cp = endp;
	      while (isspace (cp[0]))
		++cp;

	      /* If a string follows, this is the name of the current file.  */
	      if (cp[0] == '"')
		{
		  endp = ++cp;
		  while (endp[0] != '\0' && endp[0] != '"')
		    ++endp;

		  if (cp < endp)
		    {
		      /* We really found a file name.  */
		      char *new_name = (char *) xmalloc (endp - cp + 1);
		      memcpy (new_name, cp, endp - cp);
		      new_name[endp - cp] = '\0';
		      line.file_name = new_name;
		    }
		}

	      /* ... and read the next line.  */
	      continue;
	    }

	  if (isalnum (cp[6]) || cp[6] == '_'
	      || strncmp (cp, "define", 6) != 0)
	    /* Ignore line.  */
	    continue;

	  cp += 6;
	  while (isspace (*cp))
	    ++cp;
	}

      while (*cp != '\0')
	{
	  if (*cp == '"')
	    {
	      process_c_string (c_file, &cp);
	      if (extract_all != 0
		  && string.buf != NULL && string.buf[0] != '\0')
		add_id_str (NULL, string.buf, string.file_name, string.number);
	    }
	  else if (*cp == '\'')
	    skip_c_char (&cp);
	  else if (*cp == '/' && (cp[1] == '*'
				  || (cplusplus_comments & cp[1] == '/')))
	    process_comment (c_file, &cp);
	  else if (extract_all == 0 && (isalpha (*cp) || *cp == '_'))
	    {
	      /* We have to look for the standard keywords, if not
		 disabled.  */
	      if (no_default_keywords == 0 && !isalnum (cp[7])
		  && cp[7] != '_' && strncmp (cp, "gettext", 7) == 0)
		process_gettext (c_file, &cp, 0);
	      else if (no_default_keywords == 0 && !isalnum (cp[8])
		       && cp[8] != '_' && strncmp (cp, "dgettext", 8) == 0)
		process_gettext (c_file, &cp, 1);
	      else if (no_default_keywords == 0 && !isalnum (cp[9])
		       && cp[9] != '_' && strncmp (cp, "dcgettext", 9) == 0)
		process_gettext (c_file, &cp, 1);
	      else
		{
		  /* Look for the user-given keywords.  */
		  size_t cnt;
		  for (cnt = 0; cnt < nkeywords; ++cnt)
		    if (!isalnum (cp[keywords[cnt].len])
			&& cp[keywords[cnt].len] != '_'
			&& strncmp (cp, keywords[cnt].str,
				    keywords[cnt].len) == 0)
		      break;

		  if (cnt < nkeywords)
		    /* We restrict user given keywords to have the string
		       in the first argument.  */
		    process_gettext (c_file, &cp, 0);
		  else
		    /* Just a normal identifier: skip it.  */
		    do
		      ++cp;
		    while (*cp == '_' || isalnum (*cp));
		}
	    }
	  else
	    ++cp;
	}
    }
}


static void
skip_c_char (cp)
     char **cp;
{
  ++(*cp);
  while (**cp != '\0' && **cp != '\n' && **cp != '\'')
    if (*(*cp)++ == '\\')
      if (*cp != '\0' && **cp != '\n')
	++(*cp);
  if (**cp == '\'')
    ++(*cp);
}

static void
process_c_string (c_file, cp)
     FILE *c_file;
     char **cp;
{
  string.act_size = 0;

#define NEW_STR_CHAR(ch)						    \
  do {									    \
    if (string.act_size >= string.max_size)				    \
      string.buf = xrealloc (string.buf,				    \
			     (string.max_size =				    \
				  2 * (string.max_size + 100)));	    \
    string.buf[string.act_size++] = (ch);				    \
  } while (0)

  /* Get line number of first character in the string.  */
  string.number = line.number;
  string.file_name = line.file_name;

  while (**cp == '"')
    {
      /* Skip the " character.  */
      ++(*cp);

      while (**cp != '\0' && **cp != '"')
	if (**cp == '\\')
	  switch (*++(*cp))
	    {
	    case '\n':
	      {
		/* The string will continue at the beginning of the
		   next line.  */
		int len = getline (&line.buf, &line.max_size, c_file);
		if (len < 0)
		  {
		    line.buf[0] = '\0';
		    *cp = line.buf;
		    goto return_string;
		  }
		++line.number;
		line.buf[len] = '\0';

		line.act_size = (size_t) len;
		*cp = line.buf;
	      }
	      break;
	    case '\0':
	      /* The file ends while scanning a string.  */
	      goto return_string;
	    default:
	      NEW_STR_CHAR ('\\');
	      NEW_STR_CHAR (**cp);
	      if (**cp == 'n')
		{
		  NEW_STR_CHAR ('"');
		  NEW_STR_CHAR ('\n');
		  NEW_STR_CHAR ('"');
		}
	      ++(*cp);
	      break;
	    }
	else if (**cp == '\n')
	  /* This is a non-ANSI kind of multi-line strings.  */
	  {
	    int len;

	    NEW_STR_CHAR ('"');
	    NEW_STR_CHAR ('\n');
	    NEW_STR_CHAR ('"');

	    len = getline (&line.buf, &line.max_size, c_file);
	    if (len < 0)
	      {
		line.buf[0] = '\0';
		*cp = line.buf;
		goto return_string;
	      }
	    ++line.number;
	    line.buf[len] = '\0';
	    line.act_size = (size_t) len;
	    *cp = line.buf;
	  }
	else
	  NEW_STR_CHAR (*(*cp)++);

      /* Skip terminating " character.  */
      if (**cp == '"')
	++(*cp);

      /* ANSI C permits string to be concated by writing them
	 next to another.  */
      while (1)
	{
	while (isspace (**cp))
	  ++(*cp);

	if (**cp == '\0')
	  {
	    int len = getline (&line.buf, &line.max_size, c_file);
	    if (len < 0)
	      goto return_string;

	    ++line.number;
	    line.buf[len] = '\0';
	    line.act_size = (size_t) len;
	    *cp = line.buf;
	  }
	else
	  break;
	}
    }
return_string:
  /* Often the last character in a message is a '\n'.  We don't want
     this to break the line as all other appearences do.  */
  if (string.act_size >= 3
      && string.buf[string.act_size - 1] == '"'
      && string.buf[string.act_size - 2] == '\n'
      && string.buf[string.act_size - 3] == '"')
    string.act_size -= 3;

  NEW_STR_CHAR('\0');
}

static void
process_comment (c_file, cp)
     FILE *c_file;
     char **cp;
{
  /* Read a complete C (and perhaps C++) string.  We have to take care
     whether the beginning of the comment matches the argument to the
     -c command line option.  */
  int found_tag;
  char is_cplusplus = *++(*cp) == '/';

  /* Skip leading white space.  */
  while (isspace (*++(*cp)));

  /* Test whether equal to required tag.  */
  if (comment_tag != NULL
      && strncmp (comment_tag, *cp, strlen (comment_tag)) == 0)
    {
      found_tag = 1;
      *cp += strlen (comment_tag);
    }
  else
    found_tag = 0;

  /* If we don't look for tagged comments only the last comment found
     is of interest.  Discard the others.  */
  if (add_all_comments)
    comment.act_size = 0;

  if (is_cplusplus)
    {
      int len;

      /* C++ style comment.  */
      if (found_tag || add_all_comments)
	/* It is possible that we are expected to write out all comments
	   preceding keywords.  */
	{
	  size_t len = strlen (*cp);

	  if (comment.max_size <= comment.act_size + len + 7)
	    {
	      comment.max_size = 2 * MAX (comment.act_size + len + 7,
					  comment.max_size);
	      comment.buf = (char *) xrealloc (comment.buf, comment.max_size);
	    }
	  stpcpy (stpcpy (stpcpy (&comment.buf[comment.act_size],
				  "#. "),
			  *cp),
		  "#.\n");

	  comment.act_size += len + (strict_uniforum ? 6 : 4);
	  comment.number = line.number;
	}

      len = getline (&line.buf, &line.max_size, c_file);
      if (len < 0)
	len = 0;
      ++line.number;
      line.buf[len] = '\0';
      line.act_size = (size_t) len;
      *cp = line.buf;
    }
  else
    {
      /* C style comment.  */
      char *comment_end;

      /* Take care that there is enough room for the leading line.  */
      if (comment.max_size < comment.act_size + 4)
	{
	  comment.max_size = 2 * MAX (comment.max_size, 100);
	  comment.buf = (char *) xrealloc (comment.buf, comment.max_size);
	}

      /* Find possible string termination.  */
      comment_end = strchr (*cp, '*');

      while (comment_end == NULL || comment_end[1] != '/')
	{
	  if (found_tag || add_all_comments)
	    /* We have to take over the already found comment to the .po file.
	       The both possibilities in which these can happen are mutally
	       exclusive.  */
	    {
	      size_t len = comment_end ? comment_end - *cp + 1 : strlen (*cp);

	      if (comment.max_size <= comment.act_size + len + 2)
		{
		  comment.max_size = MAX (2 * (comment.act_size + len + 2),
					  2 * comment.max_size);
		  comment.buf = (char *) xrealloc (comment.buf,
						   comment.max_size);
		}
	      stpcpy (stpcpy (&comment.buf[comment.act_size], "#. "), *cp);
	      comment.act_size += len + 3;
	    }

	  if (comment_end != NULL)
	    *cp = comment_end + 1;

	  if (comment_end == NULL || **cp == '\0')
	    {
	      int len = getline (&line.buf, &line.max_size, c_file);
	      if (len < 0)
		len = 0;

	      ++line.number;
	      line.buf[len] = '\0';
	      line.act_size = (size_t) len;
	      *cp = line.buf;
	    }

	  /* Make the possible output look prettier by removing leading
	     white spaces.  */
	  while (isspace (**cp))
	    ++(*cp);

	  comment_end = strchr (*cp, '*');
	}

      if (found_tag || add_all_comments)
	{
	  size_t len = comment_end - *cp;

	  if (comment.max_size <= comment.act_size + len + 4)
	    {
	      comment.max_size = MAX (2 * (comment.act_size + len + 4),
				      2 * comment.max_size);
	      comment.buf = (char *) xrealloc (comment.buf,
					       comment.max_size);
	    }
	  memcpy (&comment.buf[comment.act_size], "#. ", 3);
	  memcpy (&comment.buf[comment.act_size + 3], *cp, len);
	  strcpy (&comment.buf[comment.act_size + len + 3], "\n#.\n");
	  comment.act_size += len + 7;

	  /* Remember line number of the string.  */
	  comment.number = line.number;
	}
      *cp = comment_end + 2;
    }
}

static void
process_gettext (c_file, cp, arg_no)
     FILE *c_file;
     char **cp;
     int arg_no;
{
  /* On entry *CP points at the beginning of the function name.  */
  char *domain = NULL;

  while (isalnum (**cp) || **cp == '_')
    ++(*cp);

#define SKIP_WS								    \
  do									    \
    {									    \
      while (isspace (**cp))						    \
	++(*cp);							    \
									    \
      if (**cp == '\0')							    \
	{								    \
	  int len = getline (&line.buf, &line.max_size, c_file);	    \
	  if (len < 0)							    \
	    len = 0;							    \
	  ++line.number;						    \
	  line.buf[len] = '\0';						    \
	  line.act_size = (size_t) len;					    \
	  *cp = line.buf;						    \
	}								    \
      else								    \
	break;								    \
    }									    \
  while (1)
  SKIP_WS;

  if (**cp != '(')
    return;
  /* Skip the ( character.  */
  ++(*cp);

  /* Skip ARG_NO parameters until the desired parameter is reached.  */
  while (arg_no-- > 0)
    {
      SKIP_WS;

      if (**cp != '"')
	/* Needless to try more.  We don't have a string as the domain.  */
	return;

      /* Extract the argument.  Result will be in STRING.BUF.  */
      process_c_string (c_file, cp);
      if (string.buf != NULL && string.buf[0] != '\0')
	domain = xstrdup (string.buf);
      else
	domain = NULL;

      SKIP_WS;

      if (**cp != ',')
	/* Needless to try more.  We don't have only a string as the domain. */
	return;
    }

  SKIP_WS;

  if (**cp != '"')
    /* This can also happen: the argument is no real string so we don't
       have anything to extract.  */
    return;

  process_c_string (c_file, cp);

  /* And add to the appropriate file/table.  */
  if (string.buf != NULL && string.buf[0] != '\0')
    add_id_str (domain != NULL ? new_domain (&new_symbols, domain) : NULL,
		string.buf, string.file_name, string.number);

  SKIP_WS;

  if (**cp != ')' && **cp != ',')
    error_with_loc (0, 0, line.file_name, line.number,
		    _("trailing garbage after string in call of keyword"));
}

/* Read the content of the specified .po file into the given structure.  */
static struct msg_domain *
read_po_file (file_name, msgs, force_domain, use_path)
     const char *file_name;
     struct msg_domain *msgs;
     const char *force_domain;
     int use_path;
{
  char *po_file_name;
  FILE *po_file = open_po_file (file_name, &po_file_name, use_path);
  char *msgid_string = NULL;
  struct msg_domain *current = NULL;

  line.number = 0;

  if (po_file == NULL)
    {
      error (0, errno, _("error while opening \"%s\" for reading"), file_name);
      return msgs;
    }

  if (force_domain != NULL)
    current = new_domain (&msgs, force_domain);

  while (!feof (po_file))
    {
      char *cp;
      char *startp, *directive;

      int len = getline (&line.buf, &line.max_size, po_file);

      if (len < 0)
	/* End of file is now reached.  */
	break;

      if (line.buf == NULL)
	error (EXIT_FAILURE, errno, _("error while reading \"%s\""),
	       po_file_name);

      /* Increase line number counter.  */
      ++line.number;
      /* And mark end of line.  */
      line.buf[len] = '\0';
      line.act_size = (size_t) len;
      cp = line.buf;

      /* Skip leading white spaces.  */
      while (isspace (*cp))
	++cp;

      /* Check for comment line.  */
      if (*cp == '#')
	{
	  /* A comment line might span across many lines when terminated
	     by "\\\n".  */
	  while (line.buf[line.act_size - 2] == '\\'
		 && line.buf[line.act_size - 1] == '\n')
	    {
	      int len = getline (&line.buf, &line.max_size, po_file);
	      if (len < 0)
		break;
	      ++line.number;
	    }
	  continue;
	}

      /* Now it's clear: the line has to be a directive.  These are
	 seperated by a white space from the rest of the line.  */
      startp = cp;
      while (*cp != '\0' && !isspace (*cp))
	++cp;
      directive = (char *) xmalloc (cp - startp + 1);
      memcpy (directive, startp, cp - startp);
      directive[cp - startp] = '\0';

      while (1)
	{
	  int len;

	  /* Skip seperating white spaces.  */
	  while (isspace (*cp))
	    ++cp;

	  if (*cp != '\\' && *(cp + 1) != '\n')
	    break;

	  len = getline (&line.buf, &line.max_size, po_file);
	  if (len < 0)
	    {
	      error (0, 0, _("incomplete `%s' directive at end of file"),
		     directive);
	      return msgs;
	    }
	  line.buf[len] = '\0';
	  line.act_size = (size_t) len;
	  cp = line.buf;
	}

      /* If nothing else follows it's an error.  */
      if (*cp == '\0')
	{
	  error_with_loc (0, 0, po_file_name, line.number,
			  _("syntax error: no value"));
	  continue;
	}

      /* Get the argument string.  The argument may extend on more than
	  one line, following the same syntax as strings in C.  */
      if (*cp != '"')
	{
	  error_with_loc (0, 0, po_file_name, line.number,
			  _("syntax error: value is no quoted string"));
	  continue;
	}
      string.act_size = 0;
      do
	{
	  /* Skip " character.  */
	  ++cp;

	  while (*cp != '\0' && *cp != '"')
	    {
	      if (string.act_size + 10 >= string.max_size)
		/* + 10 is safety margin */
		{
		  string.max_size = 2 * (string.max_size + 100);
		  string.buf = xrealloc (string.buf, string.max_size);
		}

	      string.buf[string.act_size++] = *cp;
	      if (*cp++ == '\\')
		switch (*cp)
		{
		case '\n':
		  {
		    int len;

		    string.buf[string.act_size++] = '\n';
		    len = getline (&line.buf, &line.max_size, po_file);
		    if (len < 0)
		      len = 0;
		    ++line.number;
		    line.buf[line.act_size] = '\0';
		    line.act_size = (size_t) len;
		    cp = line.buf;
		  }
		  break;
		case '\0':
		  /* Sweet nothing.  */
		  break;
		default:
		  string.buf[string.act_size++] = *cp++;
		  break;
		}
	    }

	  if (*cp == '\0')
	    error_with_loc (0, 0, po_file_name, line.number,
			    _("unterminated string constant"));
	  else
	    {
	      /* Skip " character.  */
	      ++cp;

	      while (1)
		{
		  int len;

		  while (isspace (*cp))
		    ++cp;

		  if (*cp != '\\' || *(cp + 1) != '\n')
		    break;

		  len = getline (&line.buf, &line.max_size, po_file);
		  if (len < 0)
		    len = 0;
		  else
		    ++line.number;
		  line.buf[len] = '\0';
		  line.act_size = (size_t) len;
		  cp = line.buf;
		}

	      if (*cp != '\0' && *cp != '"')
		{
		  error_with_loc (0, 0, po_file_name, line.number,
				  _("trailing garbage in line"));
		  *cp = '\0';
		}
	    }
	}
      while (*cp != '\0');
      string.buf[string.act_size] = '\0';

      if (strcmp (directive, "domain") == 0)
	{
	  if (msgid_string != NULL)
	    {
	      error_with_loc (0, 0, po_file_name, line.number,
			      _("\
previous `msgid' directive not matched by `msgstr'"));
	      msgid_string = NULL;
	    }

	  if (force_domain == NULL)
	    {
	      /* Do some test on the file name (white space etc).  */
	      size_t correct = strcspn (string.buf, INVALID_PATH_CHAR);

	      if (string.buf[correct] != '\0')
		{
		  error (0, 0, _("\
domain name \"%s\" not suitable as file name: will use prefix"),
			 string.buf);
		  string.buf[correct] = '\0';
		}

	      /* Set new domain.  */
	      current = new_domain (&msgs, string.buf);
	    }
	}
      else if (strcmp (directive, "msgstr") == 0)
	{
	  if (msgid_string == NULL)
	    error_with_loc (0, 0, po_file_name, line.number,
			    _("`msgid' directive expected"));
	  else
	    {
	      /* We found a valid pair of msgid/msgstr.  First check whether
		 already a domain is specified.  If not use default domain.  */
	      if (current == NULL)
		current = new_domain (&msgs, "messages");

	      /* We can insert the ID/string pair into the hashing table.  */
	      insert_entry (&current->symbol_tab, msgid_string,
			    strlen (msgid_string), xstrdup (string.buf));
	      free (msgid_string);
	      msgid_string = NULL;
	    }
	}
      else if (strcmp (directive, "msgid") == 0)
	{
	  if (msgid_string != NULL)
	    {
	      error_with_loc (0, 0, po_file_name, line.number,
			      _("`msgstr' directive expected"));
	      free (msgid_string);
	    }
	  msgid_string = xstrdup (string.buf);
	}
      else
	error_with_loc (0, 0, po_file_name, line.number,
			_("syntax error: unknown directive"));

      /* Now we don't need the directive name anymore.  */
      free (directive);
    }

  /* Signal no error.  */
  return msgs;
}


static struct msg_domain *
new_domain (po_content, domain_name)
     struct msg_domain **po_content;
     const char *domain_name;
{
  while (*po_content != NULL
	 && strcmp (domain_name, (*po_content)->domain_name) != 0)
    po_content = &(*po_content)->next;

  if (*po_content == NULL)
    {
      /* We must create a new domain.  */
      *po_content = (struct msg_domain *) xmalloc (sizeof (**po_content));

      if (init_hash (&(*po_content)->symbol_tab, 100) != 0)
	error (EXIT_FAILURE, errno, _("while creating hash table"));
      (*po_content)->domain_name = xstrdup (domain_name);
      (*po_content)->next = NULL;
    }

  return *po_content;
}


static void
add_id_str (dom, msgid, c_file_name, line_no)
     struct msg_domain *dom;
     const char *msgid;
     const char *c_file_name;
     size_t line_no;
{
  struct id_str *data;
  size_t id_len = strlen (msgid);
  char *msgstr;
  void *dummy;

  if (dom == NULL || extract_all)
    /* Use default domain.  */
    dom = new_symbols;

  /* Test whether the MSGID is among those from the exclude file.  */
  if (find_entry (&exclude_symbols->symbol_tab, msgid, strlen (msgid), &dummy))
    /* Simply ignore this MSGID.  */
    return;

  /* Now it is clear that we have a new entry.  Check for length limit we
     have on some systems.  */
  if (id_len > warn_id_len && verbose != 0)
    error_with_loc (0, 0, c_file_name, line_no, _("\
string length perhaps conflicts with %u byte limit"), warn_id_len);

  if (join_existing)
    {
      char *fname;

      asprintf (&fname, "%s%s.po", output_dir, dom->domain_name);
      if (fname == NULL)
	error (EXIT_FAILURE, errno, _("while preparing output"));

      /* Read existing symbols from file.  */
      read_po_file (fname, dom, dom->domain_name, 0);

      free (fname);
    }

  /* We must only initialize the msgstr entries when -m or -M were given.
     In main these variables were canonized so that both are NULL or none.  */
  if (msgstr_prefix != NULL)
    {
      /* FIXME: use alloca?  :*/
      asprintf (&msgstr, "%s%s%s", msgstr_prefix, msgid, msgstr_suffix);
      if (msgstr == NULL)
	error (EXIT_FAILURE, errno, _("while preparing output"));
    }
  else
    msgstr = NULL;

  /* Store all information in the hashing table.  */
  data = (struct id_str *) xmalloc (sizeof (*data));

  data->msgstr = msgstr;
  data->occurence.file_name = c_file_name;
  data->occurence.line_no = line_no;
  data->occurence.comment = comment.act_size > 0 ? xstrdup (comment.buf)
						 : NULL;
  data->occurence.next = NULL;

  if (insert_entry (&dom->symbol_tab, msgid, id_len, data))
    {
      /* Duplicate message ID.  No need to store this data set
	 but the line number is interesting.  */
      struct id_str *entry;
      struct file_name_line *new_line;

      (void) find_entry (&dom->symbol_tab, msgid, id_len, (void **) &entry);
      /* Get end of occurence list.  */
      new_line = &entry->occurence;
      while (new_line->next != NULL)
	new_line = new_line->next;

      /* Append new occurence.  */
      new_line->next = (struct file_name_line *) xmalloc (sizeof (*new_line));
      new_line->next->file_name = c_file_name;
      new_line->next->line_no = line_no;
      new_line->next->comment = data->occurence.comment;
      new_line->next->next = NULL;

      if (msgstr != NULL)
	free (msgstr);
      free (data);
    }

  /* Mark the comment as used.  */
  comment.act_size = 0;
}

static void
write_out_domain (dom, do_sort)
     struct msg_domain *dom;
     int do_sort;
{
  struct id_str **msg_arr = (struct id_str **) alloca (dom->symbol_tab.filled
						       * sizeof (msg_arr[0]));
  void *ptr = NULL;
  char *msgid;
  size_t cnt = 0;
  char *file_name;
  FILE *fp;

  while (iterate_table (&dom->symbol_tab, &ptr, (void **) &msgid,
			(void **) &msg_arr[cnt]))
    {
      msg_arr[cnt]->msgid = msgid;
      ++cnt;
    }

  if (do_sort)
    /* Sort the table according to the msgid.  */
    qsort (msg_arr, dom->symbol_tab.filled, sizeof (msg_arr[0]),
	   compare_msgid);

  /* FIXME: use alloca?  */
  asprintf (&file_name, "%s%s.po", output_dir, dom->domain_name);
  if (file_name == NULL)
    error (EXIT_FAILURE, errno, _("while preparing output"));

  fp = fopen (file_name, "w");
  if (fp == NULL)
    {
      error (0, errno, _("cannot create output file \"%s\""), file_name);
      free (file_name);
      return;
    }

  /* If we are just about to write to the default domain prepend a header.  */
  if (omit_header == 0 && strcmp (dom->domain_name, default_domain) == 0)
    write_header (fp);

  for (cnt = 0; cnt < dom->symbol_tab.filled; ++cnt)
    {
      /* Write out all gather comments.  */
      struct file_name_line *act_line = &msg_arr[cnt]->occurence;

      /* Begin with an empty line.  */
      fputs (strict_uniforum ? "#:\n" : "\n", fp);

      /* Process all comments collected for this message.  */
      do
	{
	  if (act_line->comment != NULL)
	    {
	      fputs (act_line->comment, fp);
	      free (act_line->comment);
	    }
	  act_line = act_line->next;
	}
      while (act_line != NULL);

      /* Write out comment with file names and lines if required.  */
      if (line_comment)
	{
	  act_line = &msg_arr[cnt]->occurence;
	  fputs ("#:", fp);
	  do
	    {
	      const char *cp = act_line->file_name;

	      while (cp[0] == '.' && cp[1] == '/')
		cp += 2;

	      fprintf (fp, " %s:%d", cp, act_line->line_no);
	      act_line = act_line->next;
	    }
	  while (act_line != NULL);
	  fputs ("\n", fp);
	}

      fprintf (fp, "msgid %s\"%s\"\nmsgstr %s\"%s\"\n",
	       strchr (msg_arr[cnt]->msgid, '\n') != NULL ? "\"\"\n" : "",
	       msg_arr[cnt]->msgid,
	       msg_arr[cnt]->msgstr != NULL
	       && strchr (msg_arr[cnt]->msgstr, '\n') != NULL ? "\"\"\n"
							      : "",
	       msg_arr[cnt]->msgstr == NULL ? "" : msg_arr[cnt]->msgstr);
    }

  /* Test for an error.  */
  if (ferror (fp))
    error (EXIT_FAILURE, 0, _("error while writing .po file"));

  /* Close output file.  */
  fclose (fp);
}

static int
compare_msgid (arg1, arg2)
     const void *arg1;
     const void *arg2;
{
  return strcmp ((*(struct id_str **)arg1)->msgid,
		 (*(struct id_str **)arg2)->msgid);
}

#define TM_YEAR_ORIGIN 1900

/* Yield A - B, measured in seconds.  */
static long
difftm (a, b)
     struct tm *a;
     struct tm *b;
{
  int ay = a->tm_year + (TM_YEAR_ORIGIN - 1);
  int by = b->tm_year + (TM_YEAR_ORIGIN - 1);
  /* Some compilers cannot handle this as a single return statement.  */
  long days = (
	       /* difference in day of year  */
	       a->tm_yday - b->tm_yday
	       /* + intervening leap days  */
	       + ((ay >> 2) - (by >> 2))
	       - (ay / 100 - by / 100)
	       + ((ay / 100 >> 2) - (by / 100 >> 2))
	       /* + diffference in years * 365  */
	       + (long) (ay - by) * 365l);

  return 60l * (60l * (24l * days + (a->tm_hour - b->tm_hour))
		+ (a->tm_min - b->tm_min))
	 + (a->tm_sec - b->tm_sec);
}

static void
write_header (output_file)
     FILE *output_file;
{
  int first = 1;
  time_t now;
  struct tm local_time;
  struct passwd *pwd;
  const char *full_user_name;
  const char *cp, *endp;
  char tz_sign;
  long tz_min;

  time (&now);
  local_time = *localtime (&now);
  tz_sign = '+';
  tz_min = difftm (&local_time, gmtime (&now)) / 60;
  if (tz_min < 0)
    {
      tz_min = -tz_min;
      tz_sign = '-';
    }

  pwd = getpwuid (getuid ());
  if (pwd == NULL || pwd->pw_gecos == NULL)
    full_user_name = "???";
  else
    full_user_name = pwd->pw_gecos;

  fprintf (output_file, "\
msgid \"\"\n\
msgstr \"\"\n\
\"Date: %d-%02d-%02d %02d:%02d:%02d%c%02ld%02ld\\n\"\n\
\"From: %s <%s>\\n\"\n\
\"Content-Type: text/plain; charset=\\n\"\n\
\"Xgettext-Options:%s\\n\"\n",
	   local_time.tm_year + TM_YEAR_ORIGIN,
	   local_time.tm_mon + 1,
	   local_time.tm_mday,
	   local_time.tm_hour,
	   local_time.tm_min,
	   local_time.tm_sec,
	   tz_sign,
	   tz_min / 60l,
	   tz_min % 60l,
	   full_user_name, get_submitter (NULL),
	   xgettext_options);

  for (cp = input_file_names; (endp = strchr (cp, '\n')) != NULL; cp = endp + 1)
    {
      fprintf (output_file, "\"%s%.*s\\n\"\n",
	       first ? "Files:" : "\t", endp - cp, cp);
      first = 0;
    }

  fprintf (output_file, "\"%s%s\\n\"\n", first ? "Files:" : "\t", cp);
}
